@@ -179,6 +179,7 @@ struct efa_cq {
 	u16 cq_idx;
 	/* NULL when no interrupts requested */
 	struct efa_eq *eq;
+	struct ib_umem *umem;
 #ifdef HAVE_EFA_KVERBS
 	u8 *buf;
 	size_t buf_size;
@@ -330,6 +331,12 @@ int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
 #else
 		  struct uverbs_attr_bundle *attrs);
 #endif
+int efa_create_cq_umem(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
+#ifndef HAVE_CREATE_CQ_BUNDLE
+		       struct ib_umem *umem, struct ib_udata *udata);
+#else
+		       struct ib_umem *umem, struct uverbs_attr_bundle *attrs);
+#endif
 #ifndef HAVE_CQ_CORE_ALLOCATION
 #ifdef HAVE_CREATE_CQ_NO_UCONTEXT
 struct ib_cq *efa_kzalloc_cq(struct ib_device *ibdev,
@@ -344,15 +351,21 @@ struct ib_cq *efa_kzalloc_cq(struct ib_device *ibdev,
 #endif
 struct ib_mr *efa_reg_mr(struct ib_pd *ibpd, u64 start, u64 length,
 			 u64 virt_addr, int access_flags,
+#ifdef HAVE_REG_USER_MR_DMAH
+			 struct ib_dmah *dmah,
+#endif
 			 struct ib_udata *udata);
 #ifdef HAVE_MR_DMABUF
 struct ib_mr *efa_reg_user_mr_dmabuf(struct ib_pd *ibpd, u64 start,
 				     u64 length, u64 virt_addr,
 				     int fd, int access_flags,
-#ifndef HAVE_REG_USER_MR_DMABUF_BUNDLE
-				     struct ib_udata *udata);
-#else
+#if defined(HAVE_REG_USER_MR_DMAH)
+				     struct ib_dmah *dmah,
 				     struct uverbs_attr_bundle *attrs);
+#elif defined(HAVE_REG_USER_MR_DMABUF_BUNDLE)
+				     struct uverbs_attr_bundle *attrs);
+#else
+				     struct ib_udata *udata);
 #endif
 #endif
 #ifdef HAVE_EFA_KVERBS
