@@ -85,6 +85,11 @@ struct efa_user_mmap_entry {
 	op(EFA_RDMA_WRITE_BYTES, "rdma_write_bytes") \
 	op(EFA_RDMA_WRITE_WR_ERR, "rdma_write_wr_err") \
 	op(EFA_RDMA_WRITE_RECV_BYTES, "rdma_write_recv_bytes") \
+	op(EFA_RETRANS_BYTES, "retrans_bytes") \
+	op(EFA_RETRANS_PKTS, "retrans_pkts") \
+	op(EFA_RETRANS_TIMEOUT_EVENS, "retrans_timeout_events") \
+	op(EFA_UNRESPONSIVE_REMOTE_EVENTS, "unresponsive_remote_events") \
+	op(EFA_IMPAIRED_REMOTE_CONN_EVENTS, "impaired_remote_conn_events") \
 
 #define EFA_STATS_ENUM(ename, name) ename,
 #ifdef HAVE_STAT_DESC_STRUCT
@@ -365,6 +370,9 @@ int efa_query_device(struct ib_device *ibdev,
 		resp.max_rdma_size = dev_attr->max_rdma_size;
 
 		resp.device_caps |= EFA_QUERY_DEVICE_CAPS_CQ_WITH_SGID;
+#if defined(HAVE_UVERBS_ATTR_RAW_FD) && defined(HAVE_IB_UMEM_DMABUF_PINNED)
+		resp.device_caps |= EFA_QUERY_DEVICE_CAPS_CQ_WITH_EXT_MEM;
+#endif
 		if (EFA_DEV_CAP(dev, RDMA_READ))
 			resp.device_caps |= EFA_QUERY_DEVICE_CAPS_RDMA_READ;
 
@@ -1138,11 +1146,11 @@ static int efa_create_qp_kernel(struct ib_qp *ibqp, struct ib_qp_init_attr *init
 
 	create_qp_params.send_cq_idx = send_cq->cq_idx;
 	create_qp_params.recv_cq_idx = recv_cq->cq_idx;
-	create_qp_params.sq_depth = init_attr->cap.max_send_wr;
-	create_qp_params.sq_ring_size_in_bytes = (init_attr->cap.max_send_wr) *
+	create_qp_params.sq_depth = qp->sq.wq.max_wqes;
+	create_qp_params.sq_ring_size_in_bytes = (qp->sq.wq.queue_mask + 1) *
 			sizeof(struct efa_io_tx_wqe);
 
-	create_qp_params.rq_depth = init_attr->cap.max_recv_wr;
+	create_qp_params.rq_depth = qp->rq.wq.max_wqes;
 	create_qp_params.rq_ring_size_in_bytes = (qp->rq.wq.queue_mask + 1) *
 			sizeof(struct efa_io_rx_desc);
 	qp->rq_size = PAGE_ALIGN(create_qp_params.rq_ring_size_in_bytes);
@@ -1660,11 +1668,12 @@ static void efa_cq_user_mmap_entries_remove(struct efa_cq *cq)
 	rdma_user_mmap_entry_remove(cq->mmap_entry);
 }
 
-#if defined(HAVE_IB_VOID_DESTROY_CQ) || defined(HAVE_IB_INT_DESTROY_CQ)
-#ifdef HAVE_IB_INT_DESTROY_CQ
+#ifdef HAVE_IB_INT_DESTROY_CQ_UDATA
 int efa_destroy_cq(struct ib_cq *ibcq, struct ib_udata *udata)
-#else
+#elif defined(HAVE_IB_VOID_DESTROY_CQ_UDATA)
 void efa_destroy_cq(struct ib_cq *ibcq, struct ib_udata *udata)
+#else
+int efa_destroy_cq(struct ib_cq *ibcq)
 #endif
 {
 	struct efa_dev *dev = to_edev(ibcq->device);
@@ -1684,54 +1693,21 @@ void efa_destroy_cq(struct ib_cq *ibcq, struct ib_udata *udata)
 #endif
 		synchronize_irq(cq->eq->irq.irqn);
 	}
-	efa_free_mapped(dev, cq->cpu_addr, cq->dma_addr, cq->size,
-			DMA_FROM_DEVICE);
-#ifndef HAVE_CQ_CORE_ALLOCATION
-	kfree(cq);
-#endif
-#ifdef HAVE_IB_INT_DESTROY_CQ
-	return 0;
-#endif
-}
-#else
-#ifdef HAVE_DESTROY_CQ_UDATA
-int efa_destroy_cq(struct ib_cq *ibcq, struct ib_udata *udata)
-#else
-int efa_destroy_cq(struct ib_cq *ibcq)
-#endif
-{
-	struct efa_dev *dev = to_edev(ibcq->device);
-	struct efa_cq *cq = to_ecq(ibcq);
-	int err;
-
-	ibdev_dbg(&dev->ibdev,
-		  "Destroy cq[%d] virt[0x%p] freed: size[%lu], dma[%pad]\n",
-		  cq->cq_idx, cq->cpu_addr, cq->size, &cq->dma_addr);
-
-	err = efa_destroy_cq_idx(dev, cq->cq_idx);
-	if (err)
-		return err;
-
-	efa_cq_user_mmap_entries_remove(cq);
-
-	if (cq->eq) {
-#ifdef HAVE_XARRAY
-		xa_erase(&dev->cqs_xa, cq->cq_idx);
-#else
-		dev->cqs_arr[cq->cq_idx] = NULL;
-#endif
-		synchronize_irq(cq->eq->irq.irqn);
-	}
-	efa_free_mapped(dev, cq->cpu_addr, cq->dma_addr, cq->size,
-			DMA_FROM_DEVICE);
 
+	if (cq->umem)
+		ib_umem_release(cq->umem);
+	else
+		efa_free_mapped(dev, cq->cpu_addr, cq->dma_addr, cq->size, DMA_FROM_DEVICE);
 #ifdef HAVE_EFA_KVERBS
 	kfree(cq->sub_cq_arr);
 #endif
+#ifndef HAVE_CQ_CORE_ALLOCATION
 	kfree(cq);
-	return 0;
-}
 #endif
+#ifndef HAVE_IB_VOID_DESTROY_CQ_UDATA
+	return 0;
+#endif
+}
 
 static struct efa_eq *efa_vec2eq(struct efa_dev *dev, int vec)
 {
@@ -1897,14 +1873,18 @@ err_out:
 	return err;
 }
 #endif
-int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
+int efa_create_cq_umem(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
 #ifndef HAVE_CREATE_CQ_BUNDLE
-		  struct ib_udata *udata)
+		       struct ib_umem *umem, struct ib_udata *udata)
 {
 #else
-		  struct uverbs_attr_bundle *attrs)
+		       struct ib_umem *umem, struct uverbs_attr_bundle *attrs)
 {
+#ifndef HAVE_EFA_KVERBS
 	struct ib_udata *udata = &attrs->driver_udata;
+#else
+	struct ib_udata *udata = NULL;
+#endif
 #endif
 #ifdef HAVE_UDATA_TO_DRV_CONTEXT
 	struct efa_ucontext *ucontext = rdma_udata_to_drv_context(
@@ -1937,6 +1917,13 @@ int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
 	}
 
 #ifdef HAVE_EFA_KVERBS
+#ifdef HAVE_CREATE_CQ_BUNDLE
+	if (attrs) {
+		udata = &attrs->driver_udata;
+		ucontext = rdma_udata_to_drv_context(udata, struct efa_ucontext, ibucontext);
+	}
+#endif
+
 	if (!udata)
 		return efa_create_cq_kernel(ibcq, attr);
 #endif
@@ -1990,11 +1977,30 @@ int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
 
 	cq->ucontext = ucontext;
 	cq->size = PAGE_ALIGN(cmd.cq_entry_size * entries * cmd.num_sub_cqs);
-	cq->cpu_addr = efa_zalloc_mapped(dev, &cq->dma_addr, cq->size,
-					 DMA_FROM_DEVICE);
-	if (!cq->cpu_addr) {
-		err = -ENOMEM;
-		goto err_out;
+
+	if (umem) {
+		if (umem->length < cq->size) {
+			ibdev_dbg(&dev->ibdev, "External memory too small\n");
+			err = -EINVAL;
+			goto err_free_mem;
+		}
+
+		if (!ib_umem_is_contiguous(umem)) {
+			ibdev_dbg(&dev->ibdev, "Non contiguous CQ unsupported\n");
+			err = -EINVAL;
+			goto err_free_mem;
+		}
+
+		cq->cpu_addr = NULL;
+		cq->dma_addr = ib_umem_start_dma_addr(umem);
+		cq->umem = umem;
+	} else {
+		cq->cpu_addr = efa_zalloc_mapped(dev, &cq->dma_addr, cq->size,
+						 DMA_FROM_DEVICE);
+		if (!cq->cpu_addr) {
+			err = -ENOMEM;
+			goto err_out;
+		}
 	}
 
 	params.uarn = cq->ucontext->uarn;
@@ -2011,7 +2017,7 @@ int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
 
 	err = efa_com_create_cq(&dev->edev, &params, &result);
 	if (err)
-		goto err_free_mapped;
+		goto err_free_mem;
 
 	resp.db_off = result.db_off;
 	resp.cq_idx = result.cq_idx;
@@ -2019,7 +2025,9 @@ int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
 	cq->ibcq.cqe = result.actual_depth;
 	WARN_ON_ONCE(entries != result.actual_depth);
 
-	err = cq_mmap_entries_setup(dev, cq, &resp, result.db_valid);
+	if (!umem)
+		err = cq_mmap_entries_setup(dev, cq, &resp, result.db_valid);
+
 	if (err) {
 		ibdev_dbg(ibdev, "Could not setup cq[%u] mmap entries\n",
 			  cq->cq_idx);
@@ -2065,15 +2073,119 @@ err_remove_mmap:
 	efa_cq_user_mmap_entries_remove(cq);
 err_destroy_cq:
 	efa_destroy_cq_idx(dev, cq->cq_idx);
-err_free_mapped:
-	efa_free_mapped(dev, cq->cpu_addr, cq->dma_addr, cq->size,
-			DMA_FROM_DEVICE);
+err_free_mem:
+	if (umem)
+		ib_umem_release(umem);
+	else
+		efa_free_mapped(dev, cq->cpu_addr, cq->dma_addr, cq->size, DMA_FROM_DEVICE);
 
 err_out:
 	atomic64_inc(&dev->stats.create_cq_err);
 	return err;
 }
 
+#if !defined(HAVE_CREATE_CQ_UMEM) && defined(HAVE_UVERBS_ATTR_RAW_FD) && defined(HAVE_IB_UMEM_DMABUF_PINNED)
+static int efa_create_cq_umem_backport(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
+#ifdef HAVE_CREATE_CQ_BUNDLE
+					struct uverbs_attr_bundle *attrs)
+{
+#else
+					struct ib_udata *udata)
+{
+	struct uverbs_attr_bundle *attrs = NULL;
+#endif
+	struct ib_device *ibdev = ibcq->device;
+	struct ib_umem_dmabuf *umem_dmabuf;
+	struct ib_umem *umem = NULL;
+	u64 buffer_length;
+	u64 buffer_offset;
+	u64 buffer_va;
+	int buffer_fd;
+	int ret;
+
+#ifndef HAVE_CREATE_CQ_BUNDLE
+	if (udata)
+		attrs = rdma_udata_to_uverbs_attr_bundle(udata);
+#endif
+
+	if (!attrs)
+		return efa_create_cq_kernel(ibcq, attr);
+
+	if (uverbs_attr_is_valid(attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_VA)) {
+		ret = uverbs_copy_from(&buffer_va, attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_VA);
+		if (ret)
+			return ret;
+
+		ret = uverbs_copy_from(&buffer_length, attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_LENGTH);
+		if (ret)
+			return ret;
+
+		if (uverbs_attr_is_valid(attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_FD) ||
+		    uverbs_attr_is_valid(attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_OFFSET))
+			return -EINVAL;
+
+		umem = ib_umem_get(ibdev, buffer_va, buffer_length, IB_ACCESS_LOCAL_WRITE);
+		if (IS_ERR(umem))
+			return PTR_ERR(umem);
+
+	} else if (uverbs_attr_is_valid(attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_FD)) {
+		ret = uverbs_get_raw_fd(&buffer_fd, attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_FD);
+		if (ret)
+			return ret;
+
+		ret = uverbs_copy_from(&buffer_offset, attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_OFFSET);
+		if (ret)
+			return ret;
+
+		ret = uverbs_copy_from(&buffer_length, attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_LENGTH);
+		if (ret)
+			return ret;
+
+		if (uverbs_attr_is_valid(attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_VA))
+			return -EINVAL;
+
+		umem_dmabuf = ib_umem_dmabuf_get_pinned(ibdev, buffer_offset, buffer_length,
+							buffer_fd, IB_ACCESS_LOCAL_WRITE);
+		if (IS_ERR(umem_dmabuf))
+			return PTR_ERR(umem_dmabuf);
+
+		umem = &umem_dmabuf->umem;
+
+	} else if (uverbs_attr_is_valid(attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_OFFSET) ||
+		   uverbs_attr_is_valid(attrs, UVERBS_ATTR_CREATE_CQ_BUFFER_LENGTH)) {
+		return -EINVAL;
+	}
+
+#ifdef HAVE_CREATE_CQ_BUNDLE
+	return efa_create_cq_umem(ibcq, attr, umem, attrs);
+#else
+	return efa_create_cq_umem(ibcq, attr, umem, udata);
+#endif
+}
+#endif
+
+#ifdef HAVE_CREATE_CQ_BUNDLE
+int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
+		  struct uverbs_attr_bundle *attrs)
+{
+#if !defined(HAVE_CREATE_CQ_UMEM) && defined(HAVE_UVERBS_ATTR_RAW_FD) && defined(HAVE_IB_UMEM_DMABUF_PINNED)
+	return efa_create_cq_umem_backport(ibcq, attr, attrs);
+#else
+	return efa_create_cq_umem(ibcq, attr, NULL, attrs);
+#endif
+}
+#else
+int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
+		  struct ib_udata *udata)
+{
+#if !defined(HAVE_CREATE_CQ_UMEM) && defined(HAVE_UVERBS_ATTR_RAW_FD) && defined(HAVE_IB_UMEM_DMABUF_PINNED)
+	return efa_create_cq_umem_backport(ibcq, attr, udata);
+#else
+	return efa_create_cq_umem(ibcq, attr, NULL, udata);
+#endif
+}
+#endif
+
 #ifndef HAVE_CQ_CORE_ALLOCATION
 #ifdef HAVE_CREATE_CQ_NO_UCONTEXT
 struct ib_cq *efa_kzalloc_cq(struct ib_device *ibdev,
@@ -2610,9 +2722,7 @@ skip_umem_pg_sz:
 
 	mr->ibmr.lkey = result.l_key;
 	mr->ibmr.rkey = result.r_key;
-#ifdef HAVE_IB_MR_LENGTH
 	mr->ibmr.length = length;
-#endif
 #ifndef HAVE_IB_MR_TYPE
 	mr->type = IB_MR_TYPE_USER;
 #endif
@@ -2637,10 +2747,13 @@ skip_umem_pg_sz:
 struct ib_mr *efa_reg_user_mr_dmabuf(struct ib_pd *ibpd, u64 start,
 				     u64 length, u64 virt_addr,
 				     int fd, int access_flags,
-#ifndef HAVE_REG_USER_MR_DMABUF_BUNDLE
-				     struct ib_udata *udata)
-#else
+#if defined(HAVE_REG_USER_MR_DMAH)
+				     struct ib_dmah *dmah,
 				     struct uverbs_attr_bundle *attrs)
+#elif defined(HAVE_REG_USER_MR_DMABUF_BUNDLE)
+				     struct uverbs_attr_bundle *attrs)
+#else
+				     struct ib_udata *udata)
 #endif
 {
 	struct efa_dev *dev = to_edev(ibpd->device);
@@ -2648,7 +2761,14 @@ struct ib_mr *efa_reg_user_mr_dmabuf(struct ib_pd *ibpd, u64 start,
 	struct efa_mr *mr;
 	int err;
 
-#ifndef HAVE_REG_USER_MR_DMABUF_BUNDLE
+#ifdef HAVE_REG_USER_MR_DMAH
+	if (dmah) {
+		err = -EOPNOTSUPP;
+		goto err_out;
+	}
+#endif
+
+#if !defined(HAVE_REG_USER_MR_DMAH) && !defined(HAVE_REG_USER_MR_DMABUF_BUNDLE)
 	mr = efa_alloc_mr(ibpd, access_flags, udata);
 #else
 	mr = efa_alloc_mr(ibpd, access_flags, &attrs->driver_udata);
@@ -2690,12 +2810,22 @@ err_out:
 
 struct ib_mr *efa_reg_mr(struct ib_pd *ibpd, u64 start, u64 length,
 			 u64 virt_addr, int access_flags,
+#ifdef HAVE_REG_USER_MR_DMAH
+			 struct ib_dmah *dmah,
+#endif
 			 struct ib_udata *udata)
 {
 	struct efa_dev *dev = to_edev(ibpd->device);
 	struct efa_mr *mr;
 	int err;
 
+#ifdef HAVE_REG_USER_MR_DMAH
+	if (dmah) {
+		err = -EOPNOTSUPP;
+		goto err_out;
+	}
+#endif
+
 	mr = efa_alloc_mr(ibpd, access_flags, udata);
 	if (IS_ERR(mr)) {
 		err = PTR_ERR(mr);
@@ -2796,9 +2926,7 @@ struct ib_mr *efa_get_dma_mr(struct ib_pd *ibpd, int access_flags)
 
 	mr->ibmr.lkey = result.l_key;
 	mr->ibmr.rkey = result.r_key;
-#ifdef HAVE_IB_MR_LENGTH
 	mr->ibmr.length = length;
-#endif
 #ifndef HAVE_IB_MR_TYPE
 	mr->type = IB_MR_TYPE_DMA;
 #endif
@@ -3516,6 +3644,7 @@ static int efa_fill_port_stats(struct efa_dev *dev, struct rdma_hw_stats *stats,
 	struct efa_com_rdma_write_stats *rws;
 	struct efa_com_rdma_read_stats *rrs;
 	struct efa_com_messages_stats *ms;
+	struct efa_com_network_stats *ns;
 	struct efa_com_basic_stats *bs;
 	int err;
 
@@ -3568,6 +3697,18 @@ static int efa_fill_port_stats(struct efa_dev *dev, struct rdma_hw_stats *stats,
 		stats->value[EFA_RDMA_WRITE_RECV_BYTES] = rws->write_recv_bytes;
 	}
 
+	params.type = EFA_ADMIN_GET_STATS_TYPE_NETWORK;
+	err = efa_com_get_stats(&dev->edev, &params, &result);
+	if (err)
+		return err;
+
+	ns = &result.network_stats;
+	stats->value[EFA_RETRANS_BYTES] = ns->retrans_bytes;
+	stats->value[EFA_RETRANS_PKTS] = ns->retrans_pkts;
+	stats->value[EFA_RETRANS_TIMEOUT_EVENS] = ns->retrans_timeout_events;
+	stats->value[EFA_UNRESPONSIVE_REMOTE_EVENTS] = ns->unresponsive_remote_events;
+	stats->value[EFA_IMPAIRED_REMOTE_CONN_EVENTS] = ns->impaired_remote_conn_events;
+
 	return ARRAY_SIZE(efa_port_stats_descs);
 }
 
@@ -3609,9 +3750,29 @@ ADD_UVERBS_METHODS(efa_mr,
 		   UVERBS_OBJECT_MR,
 		   &UVERBS_METHOD(EFA_IB_METHOD_MR_QUERY));
 
+#if !defined(HAVE_CREATE_CQ_UMEM) && defined(HAVE_UVERBS_ATTR_RAW_FD) && defined(HAVE_IB_UMEM_DMABUF_PINNED)
+ADD_UVERBS_ATTRIBUTES_SIMPLE(efa_cq_create,
+			     UVERBS_OBJECT_CQ,
+			     UVERBS_METHOD_CQ_CREATE,
+			     UVERBS_ATTR_PTR_IN(UVERBS_ATTR_CREATE_CQ_BUFFER_VA,
+						UVERBS_ATTR_TYPE(u64),
+						UA_OPTIONAL),
+			     UVERBS_ATTR_PTR_IN(UVERBS_ATTR_CREATE_CQ_BUFFER_LENGTH,
+						UVERBS_ATTR_TYPE(u64),
+						UA_OPTIONAL),
+			     UVERBS_ATTR_RAW_FD(UVERBS_ATTR_CREATE_CQ_BUFFER_FD,
+						UA_OPTIONAL),
+			     UVERBS_ATTR_PTR_IN(UVERBS_ATTR_CREATE_CQ_BUFFER_OFFSET,
+						UVERBS_ATTR_TYPE(u64),
+						UA_OPTIONAL));
+#endif
+
 const struct uapi_definition efa_uapi_defs[] = {
 	UAPI_DEF_CHAIN_OBJ_TREE(UVERBS_OBJECT_MR,
 				&efa_mr),
+#if !defined(HAVE_CREATE_CQ_UMEM) && defined(HAVE_UVERBS_ATTR_RAW_FD) && defined(HAVE_IB_UMEM_DMABUF_PINNED)
+	UAPI_DEF_CHAIN_OBJ_TREE(UVERBS_OBJECT_CQ, &efa_cq_create),
+#endif
 	{},
 };
 #endif
